@@ -7,6 +7,17 @@ from fontTools.misc.transform import Identity
 import math
 from collections import namedtuple
 
+try:
+    import cython
+
+    COMPILED = cython.compiled
+except (AttributeError, ImportError):
+    # if cython not installed, use mock module with no-op decorators and types
+    from fontTools.misc import cython
+
+    COMPILED = False
+
+
 Intersection = namedtuple("Intersection", ["pt", "t1", "t2"])
 
 
@@ -26,10 +37,13 @@ __all__ = [
     "splitCubic",
     "splitQuadraticAtT",
     "splitCubicAtT",
+    "splitCubicAtTC",
+    "splitCubicIntoTwoAtTC",
     "solveQuadratic",
     "solveCubic",
     "quadraticPointAtT",
     "cubicPointAtT",
+    "cubicPointAtTC",
     "linePointAtT",
     "segmentPointAtT",
     "lineLineIntersections",
@@ -67,6 +81,14 @@ def _split_cubic_into_two(p0, p1, p2, p3):
     )
 
 
+@cython.returns(cython.double)
+@cython.locals(
+    p0=cython.complex,
+    p1=cython.complex,
+    p2=cython.complex,
+    p3=cython.complex,
+)
+@cython.locals(mult=cython.double, arch=cython.double, box=cython.double)
 def _calcCubicArcLengthCRecurse(mult, p0, p1, p2, p3):
     arch = abs(p0 - p3)
     box = abs(p0 - p1) + abs(p1 - p2) + abs(p2 - p3)
@@ -79,6 +101,17 @@ def _calcCubicArcLengthCRecurse(mult, p0, p1, p2, p3):
         )
 
 
+@cython.returns(cython.double)
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+)
+@cython.locals(
+    tolerance=cython.double,
+    mult=cython.double,
+)
 def calcCubicArcLengthC(pt1, pt2, pt3, pt4, tolerance=0.005):
     """Calculates the arc length for a cubic Bezier segment.
 
@@ -97,14 +130,22 @@ epsilonDigits = 6
 epsilon = 1e-10
 
 
+@cython.cfunc
+@cython.inline
+@cython.returns(cython.double)
+@cython.locals(v1=cython.complex, v2=cython.complex)
 def _dot(v1, v2):
     return (v1 * v2.conjugate()).real
 
 
+@cython.cfunc
+@cython.inline
+@cython.returns(cython.double)
+@cython.locals(x=cython.complex)
 def _intSecAtan(x):
     # In : sympy.integrate(sp.sec(sp.atan(x)))
     # Out: x*sqrt(x**2 + 1)/2 + asinh(x)/2
-    return x * math.sqrt(x ** 2 + 1) / 2 + math.asinh(x) / 2
+    return x * math.sqrt(x**2 + 1) / 2 + math.asinh(x) / 2
 
 
 def calcQuadraticArcLength(pt1, pt2, pt3):
@@ -142,6 +183,25 @@ def calcQuadraticArcLength(pt1, pt2, pt3):
     return calcQuadraticArcLengthC(complex(*pt1), complex(*pt2), complex(*pt3))
 
 
+@cython.returns(cython.double)
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    d0=cython.complex,
+    d1=cython.complex,
+    d=cython.complex,
+    n=cython.complex,
+)
+@cython.locals(
+    scale=cython.double,
+    origDist=cython.double,
+    a=cython.double,
+    b=cython.double,
+    x0=cython.double,
+    x1=cython.double,
+    Len=cython.double,
+)
 def calcQuadraticArcLengthC(pt1, pt2, pt3):
     """Calculates the arc length for a quadratic Bezier segment.
 
@@ -154,7 +214,7 @@ def calcQuadraticArcLengthC(pt1, pt2, pt3):
         Arc length value.
     """
     # Analytical solution to the length of a quadratic bezier.
-    # I'll explain how I arrived at this later.
+    # Documentation: https://github.com/fonttools/fonttools/issues/3055
     d0 = pt2 - pt1
     d1 = pt3 - pt2
     d = d1 - d0
@@ -191,6 +251,17 @@ def approximateQuadraticArcLength(pt1, pt2, pt3):
     return approximateQuadraticArcLengthC(complex(*pt1), complex(*pt2), complex(*pt3))
 
 
+@cython.returns(cython.double)
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+)
+@cython.locals(
+    v0=cython.double,
+    v1=cython.double,
+    v2=cython.double,
+)
 def approximateQuadraticArcLengthC(pt1, pt2, pt3):
     """Calculates the arc length for a quadratic Bezier segment.
 
@@ -288,6 +359,20 @@ def approximateCubicArcLength(pt1, pt2, pt3, pt4):
     )
 
 
+@cython.returns(cython.double)
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+)
+@cython.locals(
+    v0=cython.double,
+    v1=cython.double,
+    v2=cython.double,
+    v3=cython.double,
+    v4=cython.double,
+)
 def approximateCubicArcLengthC(pt1, pt2, pt3, pt4):
     """Approximates the arc length for a cubic Bezier segment.
 
@@ -549,6 +634,70 @@ def splitCubicAtT(pt1, pt2, pt3, pt4, *ts):
     return _splitCubicAtT(a, b, c, d, *ts)
 
 
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+    a=cython.complex,
+    b=cython.complex,
+    c=cython.complex,
+    d=cython.complex,
+)
+def splitCubicAtTC(pt1, pt2, pt3, pt4, *ts):
+    """Split a cubic Bezier curve at one or more values of t.
+
+    Args:
+        pt1,pt2,pt3,pt4: Control points of the Bezier as complex numbers..
+        *ts: Positions at which to split the curve.
+
+    Yields:
+        Curve segments (each curve segment being four complex numbers).
+    """
+    a, b, c, d = calcCubicParametersC(pt1, pt2, pt3, pt4)
+    yield from _splitCubicAtTC(a, b, c, d, *ts)
+
+
+@cython.returns(cython.complex)
+@cython.locals(
+    t=cython.double,
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+    pointAtT=cython.complex,
+    off1=cython.complex,
+    off2=cython.complex,
+)
+@cython.locals(
+    t2=cython.double, _1_t=cython.double, _1_t_2=cython.double, _2_t_1_t=cython.double
+)
+def splitCubicIntoTwoAtTC(pt1, pt2, pt3, pt4, t):
+    """Split a cubic Bezier curve at t.
+
+    Args:
+        pt1,pt2,pt3,pt4: Control points of the Bezier as complex numbers.
+        t: Position at which to split the curve.
+
+    Returns:
+        A tuple of two curve segments (each curve segment being four complex numbers).
+    """
+    t2 = t * t
+    _1_t = 1 - t
+    _1_t_2 = _1_t * _1_t
+    _2_t_1_t = 2 * t * _1_t
+    pointAtT = (
+        _1_t_2 * _1_t * pt1 + 3 * (_1_t_2 * t * pt2 + _1_t * t2 * pt3) + t2 * t * pt4
+    )
+    off1 = _1_t_2 * pt1 + _2_t_1_t * pt2 + t2 * pt3
+    off2 = _1_t_2 * pt2 + _2_t_1_t * pt3 + t2 * pt4
+
+    pt2 = pt1 + (pt2 - pt1) * t
+    pt3 = pt4 + (pt3 - pt4) * _1_t
+
+    return ((pt1, pt2, off1, pointAtT), (pointAtT, off2, pt3, pt4))
+
+
 def _splitQuadraticAtT(a, b, c, *ts):
     ts = list(ts)
     segments = []
@@ -611,6 +760,44 @@ def _splitCubicAtT(a, b, c, d, *ts):
     return segments
 
 
+@cython.locals(
+    a=cython.complex,
+    b=cython.complex,
+    c=cython.complex,
+    d=cython.complex,
+    t1=cython.double,
+    t2=cython.double,
+    delta=cython.double,
+    delta_2=cython.double,
+    delta_3=cython.double,
+    a1=cython.complex,
+    b1=cython.complex,
+    c1=cython.complex,
+    d1=cython.complex,
+)
+def _splitCubicAtTC(a, b, c, d, *ts):
+    ts = list(ts)
+    ts.insert(0, 0.0)
+    ts.append(1.0)
+    for i in range(len(ts) - 1):
+        t1 = ts[i]
+        t2 = ts[i + 1]
+        delta = t2 - t1
+
+        delta_2 = delta * delta
+        delta_3 = delta * delta_2
+        t1_2 = t1 * t1
+        t1_3 = t1 * t1_2
+
+        # calc new a, b, c and d
+        a1 = a * delta_3
+        b1 = (3 * a * t1 + b) * delta_2
+        c1 = (2 * b * t1 + c + 3 * a * t1_2) * delta
+        d1 = a * t1_3 + b * t1_2 + c * t1 + d
+        pt1, pt2, pt3, pt4 = calcCubicPointsC(a1, b1, c1, d1)
+        yield (pt1, pt2, pt3, pt4)
+
+
 #
 # Equation solvers.
 #
@@ -773,6 +960,24 @@ def calcCubicParameters(pt1, pt2, pt3, pt4):
     return (ax, ay), (bx, by), (cx, cy), (dx, dy)
 
 
+@cython.cfunc
+@cython.inline
+@cython.locals(
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+    a=cython.complex,
+    b=cython.complex,
+    c=cython.complex,
+)
+def calcCubicParametersC(pt1, pt2, pt3, pt4):
+    c = (pt2 - pt1) * 3.0
+    b = (pt3 - pt2) * 3.0 - c
+    a = pt4 - pt1 - c - b
+    return (a, b, c, pt1)
+
+
 def calcQuadraticPoints(a, b, c):
     ax, ay = a
     bx, by = b
@@ -802,6 +1007,24 @@ def calcCubicPoints(a, b, c, d):
     return (x1, y1), (x2, y2), (x3, y3), (x4, y4)
 
 
+@cython.cfunc
+@cython.inline
+@cython.locals(
+    a=cython.complex,
+    b=cython.complex,
+    c=cython.complex,
+    d=cython.complex,
+    p2=cython.complex,
+    p3=cython.complex,
+    p4=cython.complex,
+)
+def calcCubicPointsC(a, b, c, d):
+    p2 = c * (1 / 3) + d
+    p3 = (b + c) * (1 / 3) + p2
+    p4 = a + b + c + d
+    return (d, p2, p3, p4)
+
+
 #
 # Point at time
 #
@@ -845,21 +1068,47 @@ def cubicPointAtT(pt1, pt2, pt3, pt4, t):
     Returns:
         A 2D tuple with the coordinates of the point.
     """
+    t2 = t * t
+    _1_t = 1 - t
+    _1_t_2 = _1_t * _1_t
     x = (
-        (1 - t) * (1 - t) * (1 - t) * pt1[0]
-        + 3 * (1 - t) * (1 - t) * t * pt2[0]
-        + 3 * (1 - t) * t * t * pt3[0]
-        + t * t * t * pt4[0]
+        _1_t_2 * _1_t * pt1[0]
+        + 3 * (_1_t_2 * t * pt2[0] + _1_t * t2 * pt3[0])
+        + t2 * t * pt4[0]
     )
     y = (
-        (1 - t) * (1 - t) * (1 - t) * pt1[1]
-        + 3 * (1 - t) * (1 - t) * t * pt2[1]
-        + 3 * (1 - t) * t * t * pt3[1]
-        + t * t * t * pt4[1]
+        _1_t_2 * _1_t * pt1[1]
+        + 3 * (_1_t_2 * t * pt2[1] + _1_t * t2 * pt3[1])
+        + t2 * t * pt4[1]
     )
     return (x, y)
 
 
+@cython.returns(cython.complex)
+@cython.locals(
+    t=cython.double,
+    pt1=cython.complex,
+    pt2=cython.complex,
+    pt3=cython.complex,
+    pt4=cython.complex,
+)
+@cython.locals(t2=cython.double, _1_t=cython.double, _1_t_2=cython.double)
+def cubicPointAtTC(pt1, pt2, pt3, pt4, t):
+    """Finds the point at time `t` on a cubic curve.
+
+    Args:
+        pt1, pt2, pt3, pt4: Coordinates of the curve as complex numbers.
+        t: The time along the curve.
+
+    Returns:
+        A complex number with the coordinates of the point.
+    """
+    t2 = t * t
+    _1_t = 1 - t
+    _1_t_2 = _1_t * _1_t
+    return _1_t_2 * _1_t * pt1 + 3 * (_1_t_2 * t * pt2 + _1_t * t2 * pt3) + t2 * t * pt4
+
+
 def segmentPointAtT(seg, t):
     if len(seg) == 2:
         return linePointAtT(*seg, t)
